/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef PORTABLE_DATA_IO_H
#define PORTABLE_DATA_IO_H

#include "support/PortableStatusConstants.h"
#include "support/PortableTypeDefs.h"
#include "support/PortableMacros.h"

/** 
 *  Abstract base class for a byte-stream Data I/O interface, similar to Be's BDataIO.
 */
class MUSCLE_CLASS_DECLARATIONS PortableDataIO
{
public:
   /** Values to pass in to PortableDataIO::Seek()'s second parameter */
   enum {
      IO_SEEK_SET = 0,
      IO_SEEK_CUR,
      IO_SEEK_END,
      NUM_IO_SEEKS
   };

   /**
    *  Virtual destructor, to keep C++ honest.
    */
   virtual ~PortableDataIO() {}

   /** Tries to place (size) bytes of new data into (buffer).  Returns the
    *  actual number of bytes placed, or a negative value if there
    *  was an error.
    *  @param buffer Buffer to write the bytes into
    *  @param size Number of bytes in the buffer.
    *  @return Number of bytes read, or -1 on error.   
    */
   virtual ssize_t Read(void * buffer, size_t size) = 0;

   /** Takes (size) bytes from (buffer) and pushes them in to the
    *  outgoing I/O stream.  Returns the actual number of bytes 
    *  read from (buffer) and pushed, or a negative value if there
    *  was an error.
    *  @param buffer Buffer to read the bytes from.
    *  @param size Number of bytes in the buffer.
    *  @return Number of bytes written, or -1 on error.        
    */
   virtual ssize_t Write(const void * buffer, size_t size) = 0;

   /**
    * New for v1.71:  Seek to a given position in the I/O stream.  
    * May not be supported by a PortableDataIO subclass, in 
    * which case B_ERROR will always be returned.
    * @param offset Byte offset to seek to or by (depending on the next arg)
    * @param whence Set this to IO_SEEK_SET if you want the offset to
    *               be relative to the start of the stream; or to 
    *               IO_SEEK_CUR if it should be relative to the current
    *               stream position, or IO_SEEK_END if it should be
    *               backwards from the end of the stream.
    * @return B_NO_ERROR on success, or B_ERROR on failure or if unsupported.
    */
   virtual status_t Seek(int64 offset, int whence) = 0;

   /** 
    * New for v1.83:  Returns the max number of microseconds to allow
    * for an output stall, before presuming that the I/O is hosed.
    * Default implementation returns ((uint64)-1), aka no limit.
    */
   virtual uint64 GetOutputStallLimit() const {return ((uint64)-1);}

   /**
    * Flushes the output buffer, if possible.  For some implementations,
    * this is a no-op.  For others (e.g. TCPSocketDataIO) this can be
    * called to improve latency of outgoing data blocks.
    */
   virtual void FlushOutput() = 0;

   /** 
    * Closes the connection.  After calling this method, the
    * PortableDataIO object should not be used any more.
    */
   virtual void Shutdown() = 0;

   /**
    * New for v1.64:  Optional interface for returning information
    * on when a given byte returned by the previous Read() call
    * was received.  Not implemented by default, and not implemented
    * by any of the standard MUSCLE PortableDataIO subclasses. (Used
    * by an LCS dataIO class that needs precision timing)
    * @param whichByte Index of the byte in the previously returned 
    *                  read-buffer that you are interested in.
    * @param retStamp On success, this value is set to the timestamp
    *                 of the byte.
    * @return B_NO_ERROR if a timestamp was written into (retStamp),
    *                    otherwise B_ERROR.  Default implementation
    *                    always returns B_ERROR.
    */
   virtual status_t GetReadByteTimeStamp(ssize_t /*whichByte*/, uint64 & /*retStamp*/) const {return B_ERROR;}

};

#endif
