/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef QMESSAGETRANSCEIVERTHREAD_H
#define QMESSAGETRANSCEIVERTHREAD_H

#include <qobject.h>
#include <qthread.h>
#include "iogateway/AbstractMessageIOGateway.h"
#include "qtsupport/QThreadSafeObjectPool.h"

/** Type representing any function that will return a newly allocated AbstractMessageIOGateway object that uses the given PortableDataIO object. */
typedef AbstractMessageIOGateway* (*GatewayFactoryFunc)(PortableDataIO * io);

/**
 *  This is a utility class to help interface Qt GUI
 *  code more easily to a TCP socket that is sending and receiving
 *  PortableMessages.   Only compiles under Qt!
 *  It helps by allowing you to send and receive
 *  PortableMessages without having to worry about calling select(),
 *  or any blocking issues.
 * 
 *  Usage:  
 *          -# Create a new QMessageTransceiverThread object.
 *          -# Call StartThread(), StartAcceptThread(), StartServerThread(), or StartConnectThread().
 *          -# Whenever you wish to send a PortableMessage over the TCP socket,
 *             simply call SendOutgoingMessage().  
 *          -# When the TCP connection is established, the Connected() signal will be emitted to notify you.
 *          -# Whenever a PortableMessage is received, the PortableMessageReceived()
 *             signal will be emitted.  Any connected components will then be able to
 *             process the incoming PortableMessage.
 *          -# When the TCP connection is broken, the Disconnected() signal will be emitted to notify you.
 *          -# When you are done, just delete this object normally.  If you wish,
 *             you can call WaitForAllMessagesToBeSent() first, in order to ensure
 *             that all pending outbound messages get delivered.
 *          -# That's it!
 */
class MUSCLE_CLASS_DECLARATIONS QMessageTransceiverThread : public QObject
{
   Q_OBJECT

public:
   /** Creates a receiver thread without a reply target.  You should
     * make all your calls to connect() before calling StartThread()...
     */
   QMessageTransceiverThread(QObject * parent = NULL, const char * name = NULL);

   /** 
    *  Destructor.  Safe to call at any time. 
    */
   virtual ~QMessageTransceiverThread();

   /** Creates a TCP send/receive thread that connects to the given hostname and port.
    *  @param hostname is the Internet hostname of the computer to connect to.
    *  @param port is the port number to connect on.
    *  @return B_NO_ERROR if the thread was started successfully, or B_ERROR if it failed.
    */
   status_t StartConnectThread(const char * hostname, int port);

   /**
    * Creates a TCP send/receive thread that waits for connections on the given port.
    *  @param port the port to listen for incoming connections on.  If 0, then the system 
    *              will choose a port to accept on.  The port number it chooses can be discovered by calling GetPort().
    *  @param acceptMultiple if true, will allow subsequent connections to be accepted after the first one is closed.
    *                        In no case will there be more than one connection active at a time, however.
    *  @return B_NO_ERROR if the thread was started successfully, or B_ERROR if it failed.
    */
   status_t StartAcceptThread(int port, bool acceptMultiple);

   /**
    * Creates a TCP accept thread that waits for connections on the given port.
    * The difference between calling this call and StartAcceptThread() lies in how the
    * incoming connection is handled.  With StartAcceptThread(), the incoming connection
    * becomes the 'current connection' for this object, and PortableMessages are sent and
    * received as usual.  With this method, on the other hand, an incoming connection causes
    * the IncomingConnectionReceived() signal to be emitted, and it is up to your code to deal
    * with the new socket. 
    *  @param port the port to listen for incoming messages on.  If set to 0, the system will 
    *         choose a port to accept on.  The port number it chooses can be discovered by calling GetPort().
    *  @param acceptMultiple if false, then only the first incoming connection will ever be accepted.
    *  @return B_NO_ERROR if the thread was started successfully, or B_ERROR if it failed.
    */
   status_t StartServerThread(int port, bool acceptMultiple=true);

   /**
    * Creates and activates the TCP send/receive background thread.
    *  @param gateway is the AbstractMessageIOGateway the thread should use to send and receive data.
    *                 In general, a AbstractMessageIOGateway object with a non-blocking TCPSocketDataIO
    *                 that uses your socket is the way to go here.
    *  @param datafd The TCP socket or file descriptor to select() on.  Assumed to be owned
    *                by (gateway), but passed in separately anyway.
    *  @return B_NO_ERROR if the thread was started successfully, or B_ERROR if it failed.
    */
   status_t StartThread(AbstractMessageIOGateway * gateway, int datafd);
   
   /**
    * Blocks until all messages that were in the outgoing queue have been sent.
    */
   void WaitForAllMessagesToBeSent();

   /**
    * Returns the port number that the accept thread is accepting messages on. 
    * This number is only meaningful if StartAcceptThread() or StartServerThread() was called.
    */
   int GetPort() const;

   /**
    * If you wish to use a type of AbstractMessageIOGateway other than the default type,
    * you can set a callback to your own factory function here.  The factory function should
    * return a new AbstractMessageIOGateway object (or subclass thereof), or NULL on failure.
    * The returned object becomes property of this QMessageTransceiverThread object.
    * @param func the new factory function this QMessageTransceiverThread should use to create its AbstractMessageIOGateway.
    */
   void SetGatewayFactoryFunc(GatewayFactoryFunc func);

   /**
    * Returns a pointer to a singleton thread-safe message pool.
    * Equivalent to ::GetMessagePool(), but for use in multithreaded Qt environments.
    */
   static QThreadSafeObjectPool<PortableMessage> * GetMessagePool() {return &_msgPool;}

signals:
   /**
    *  Signal emitted when a TCP connection is completed.
    */
   void Connected();

   /**
    *  Signal emitted when a TCP connection has been broken.
    */
   void Disconnected();

   /**
    *  Signal emitted when a new PortableMessage has arrived over the TCP connection.
    *  @param msgRef A reference to the newly received PortableMessage.
    *  @param numLeftInQueue how many more PortableMessages are still in the local incoming message
    *         queue after this one.
    */
   void PortableMessageReceived(PortableMessageRef msgRef, uint32 numLeftInQueue);

   /**
    *  Signal emitted when the outgoing PortableMessage queue has become empty.
    */
   void OutgoingMessageQueueDrained();

   /**
    *  Signal emitted when a new TCP connection has been accepted.  Used in conjunction
    *  with StartServerThread() only.
    *  @param socketFD the TCP socket descriptor of the newly accepted connection.
    *                  The receiving code is expected to assume ownership of this socket.
    */
   void IncomingConnectionReceived(int socketFD);

public slots:
   /**
    * Enqueues a PortableMessage for immediate output over the TCP connection.
    * @param msgRef a reference to the PortableMessage to send out.
    * @return B_NO_ERROR on success, B_ERROR if no more messages can be sent (e.g. connection is broken)
    * NOTE:  If you want to use object pools with your (msgRef), make sure that the
    *        pools you use are QThreadSafeObjectPools objects, or you will suffer from
    *        nasty race conditions.  The GetMessagePool() method in this class returns
    *        the appropriate pool object to use here.
    */
   status_t SendOutgoingMessage(PortableMessageRef msgRef);

   /** Kills any existing communications thread and resets the MessageTransceiverThread
    *  object to its initial state.  After this call returns, you can call one of the
    *  Start...() methods again to begin another session.  Safe to call at any time.
    *  (An alternative to deleting the MessageTransceiverThread object and creating a new one)
    */
   void Reset();

private:
   virtual bool event(QEvent * evt);

   AbstractMessageIOGateway * CreateIOGateway(int socketfd) const;
   typedef void (*InternalThreadFunc)(QMessageTransceiverThread * owner, QThread * thread);

   status_t StartThread(InternalThreadFunc func, const char * threadName);
   void NotifyTCPThread();
   status_t CreateNotificationMechanism();

   static void TCPThreadStub(QMessageTransceiverThread * This, QThread * thread) {This->TCPThreadLoop(thread);}
   static void ConnectThreadStub(QMessageTransceiverThread * This, QThread * thread) {This->ConnectThreadLoop(thread);}
   void TCPThreadLoop(QThread * thread);
   void ConnectThreadLoop(QThread * thread);

   void AcquireMessagesPendingLock();
   void ReleaseMessagesPendingLock();
   void NotifyEventThread(int eventCode);

   enum {
      QMTT_EVENT_CONNECTED = 8360446,  // an arbitrary number to avoid clashes (hopefully)
      QMTT_EVENT_DISCONNECTED,
      QMTT_EVENT_NEWMESSAGES,
      QMTT_EVENT_NEWCONNECTIONS,
      QMTT_EVENT_DRAINED,
   };

   volatile QThread * _tcpThread;
   QMutex _mutex;   

   // _gateway and _datafd are linked.  If _gateway if non-NULL, _datafd must be a valid fd.
   // And if _gateway is NULL, _datafd must be < 0.
   AbstractMessageIOGateway * _gateway;
   int _datafd;    // the fd we actually transfer messages over

   int _acceptfd;  // the fd we wait for connections on (or -1 if we aren't doing that)

   int _notifyfd;  // the fd we use to wake up our thread (hacksville!)
   int _waitfd;    // the fd the thread uses to be woken up (hack hack!)

   PortableString _connectTo;  // used by StartConnectThread()
   int _port;
   int32 _threadPri;
 
   bool _isConnecting;
   bool _isServer;      // true iff StartServerThread() was called

   bool _acceptMultiple;

   QSemaphore _messagesPending;

   GatewayFactoryFunc _userGatewayFunc;

   PortableQueue<int> _incomingConnections;  // used when in server mode 

   static QThreadSafeObjectPool<PortableMessage> _msgPool;

   // Called by our InternalMessageThread object
   void threadRun();

   // Our internal messaging thread implementation
   class InternalMessageThread : public QThread
   {
   public:
      InternalMessageThread(QMessageTransceiverThread * owner, InternalThreadFunc func) : _owner(owner), _func(func) {/* empty */}
      virtual void run() {_func(_owner, this);}

   private:
      QMessageTransceiverThread * _owner;
      InternalThreadFunc _func;
   };

   PortableQueue<PortableMessageRef> _tempIncomingQueue;  // kept around to avoid excessive buffer reallocations
};

#endif
