/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include "reflector/AbstractReflectSession.h"
#include "reflector/ReflectServer.h"
#include "iogateway/PortableMessageIOGateway.h"

static uint32 _idCounter = 0L;

AbstractReflectSession ::
AbstractReflectSession() : _owner(NULL), _socket(-1), _gateway(NULL), _lastByteOutputAt(0)
{
   sprintf(_idString, "%lu", _idCounter++);  
}

AbstractReflectSession ::
~AbstractReflectSession() 
{
   delete _gateway;
}

const char * 
AbstractReflectSession ::
GetHostName() const 
{
   MASSERT(_owner, "Can't call GetHostName() while not attached to the server");
   return _hostName.Cstr();
}

int
AbstractReflectSession ::
GetClientPort() const 
{
   MASSERT(_owner, "Can't call GetClientPort() while not attached to the server");
   return _port;
}

const char * 
AbstractReflectSession ::
GetSessionIDString() const 
{
   return _idString;
}

PortableMessage & 
AbstractReflectSession ::
GetCentralState() const 
{
   MASSERT(_owner, "Can't call GetCentralState() while not attached to the server");
   return _owner->GetCentralState();
}

status_t 
AbstractReflectSession ::
AddOutgoingMessage(PortableMessageRef ref) 
{
   MASSERT(_owner, "Can't call AddOutgoingMessage() while not attached to the server");
   return (_gateway) ? _gateway->AddOutgoingMessage(ref) : B_ERROR;
}

AbstractMessageIOGateway * 
AbstractReflectSession ::
CreateGateway(int socket)
{
   TCPSocketDataIO * io = newnothrow TCPSocketDataIO(socket, false);
   if (io)
   {
      AbstractMessageIOGateway * gw = newnothrow PortableMessageIOGateway(io);
      if (gw) return gw;
      io->ReleaseSocket();  // caller still owns the socket on error
      delete io;
   }
   WARN_OUT_OF_MEMORY;
   return NULL;
}

status_t 
AbstractReflectSession ::
AttachedToServer()
{
   return B_NO_ERROR;
}

PortableHashtableIterator<const char *, AbstractReflectSessionRef>
AbstractReflectSession ::
GetSessions() const
{
   MASSERT(_owner, "Can't call GetSessions() while not attached to the server");
   return _owner->GetSessions(); 
}

AbstractReflectSessionRef 
AbstractReflectSession ::
GetSession(const char * id) const
{
   MASSERT(_owner, "Can't call GetSession() while not attached to the server");
   return _owner->GetSession(id);
}

status_t
AbstractReflectSession ::
AddNewSession(AbstractReflectSessionRef ref, int socket)
{
   MASSERT(_owner, "Can't call AddNewSession() while not attached to the server");
   return _owner->AddNewSession(ref, socket);
}

void
AbstractReflectSession ::
EndSession()
{
   MASSERT(_owner, "Can't call EndSession() while not attached to the server");
   _owner->EndSession(this);
}

void
AbstractReflectSession ::
EndServer()
{
   MASSERT(_owner, "Can't call EndServer() while not attached to the server");
   _owner->EndServer();
}

uint64
AbstractReflectSession ::
GetServerUptime() const
{
   MASSERT(_owner, "Can't call GetServerUptime() while not attached to the server");
   return _owner->GetServerUptime();
}

status_t
AbstractReflectSession ::
PutAcceptPort(ReflectSessionFactory * factory, short port) 
{
   MASSERT(_owner, "Can't call PutAcceptPort() while not attached to the server");
   return _owner->PutAcceptPort(factory, port);
}

void
AbstractReflectSession ::
RemoveAcceptPort(short port)
{
   MASSERT(_owner, "Can't call RemoveAcceptPort() while not attached to the server");
   _owner->RemoveAcceptPort(port);
}

status_t
AbstractReflectSession ::
ReplaceSession(AbstractReflectSessionRef replaceMeWithThis)
{
   MASSERT(_owner, "Can't call ReplaceSession() while not attached to the server");
   return _owner->ReplaceSession(replaceMeWithThis, this);
}

bool
AbstractReflectSession ::
ClientConnectionClosed()
{
   return true;  // true == okay to remove this session
}

void
AbstractReflectSession ::
SetGateway(AbstractMessageIOGateway * gw, int socket, bool deleteOld)
{
   _socket = socket;
   if (deleteOld) delete _gateway;
   _gateway = gw;
}

void
AbstractReflectSession ::
BroadcastToAllNeighbors(PortableMessageRef msgRef, void * userData, bool toSelf)
{
   PortableHashtableIterator<const char *, AbstractReflectSessionRef> iter = GetSessions();
   AbstractReflectSessionRef * next;
   while((next = iter.GetNextValue()) != NULL)
   {
      AbstractReflectSession * session = next->GetItemPointer();
      if ((session)&&((toSelf)||(session != this))) session->MessageReceivedFromNeighbor(*this, msgRef, userData);
   }
}
