/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <new>
#include <typeinfo>
#include <string.h>

#include "reflector/ReflectServer.h"
#include "util/NetworkUtilityFunctions.h"

static uint32 _outOfMemoryWarningCount = 0L;
static uint32 _maxAllocableBytes = (uint32)-1;

static uint32 _bytesAlloced = 0L;

static PortableQueue<OutOfMemoryHandlerFunc> _outOfMemoryHandlers;

status_t AddOutOfMemoryHandler(OutOfMemoryHandlerFunc f)
{
   return _outOfMemoryHandlers.AddTail(f) ? B_NO_ERROR : B_ERROR;
}

uint32 GetNumAvailableBytes()
{
   return (_bytesAlloced < _maxAllocableBytes) ? _maxAllocableBytes-_bytesAlloced : 0;
}
                                                              
uint32 GetNumUsedBytes()
{
   return _bytesAlloced;
}

uint32 GetMaxNumBytes()
{
   return _maxAllocableBytes;
}

// Metrowerk's compiler crashes at run-time if this memory-tracking
// code is enabled.  I haven't been able to figure out why (everything
// looks okay, but after a few dozen malloc()/free() calls, free()
// crashes!) so I'm just going to disable this code for PowerPC/Metrowerks
// machines.  Sorry!   --jaf 12/01/00
#ifdef __MWERKS__
#define DISABLE_MUSCLE_MEMORY_TRACKING
#endif

#ifndef DISABLE_MUSCLE_MEMORY_TRACKING

static void * jallocAux(size_t);
void * jallocAux(size_t s)
{
    if (_bytesAlloced + s > _maxAllocableBytes) return NULL;

    uint32 allocSize = s + sizeof(uint32);  // requested size, plus 4 bytes for our use
    uint32 * a = (uint32*) malloc(allocSize);
    if (a)
    {
       *a = allocSize;  // a tag so jfree() will know how big the alloc was
       _bytesAlloced += allocSize;
//printf("+%li = %lu\n", allocSize, _bytesAlloced);
       return a+1;  // user doesn't see it, of course
    }
    else return NULL;
}

static void * jalloc(size_t);
void * jalloc(size_t s)
{
   void * ret = jallocAux(s);
   if (ret == NULL)
   {
      GetMessagePool()->Drain();     // dump some unnecessary objects too
      for (int i=_outOfMemoryHandlers.CountItems()-1; i>=0; i--) _outOfMemoryHandlers[i]();  // user-defined lo-mem callbacks
      ret = jallocAux(s);
   }
   if (ret == NULL) _outOfMemoryWarningCount++;  // still no mem?  Okay, time to break out the whoop ass
   return ret;
}

static void jfree(void *);
void jfree(void * p)
{
    uint32 * realStart = ((uint32*)p)-1;
    uint32 s = *realStart;
    _bytesAlloced -= s;
//printf("-%li = %lu\n", s, _bytesAlloced);
    free(realStart);
}

// VC++ doesn't know about these  :^(
#ifdef WIN32
# define bad_alloc()
# define nothrow_t
#endif

#ifdef WIN32
void * operator new(size_t s)
#else
void * operator new(size_t s) throw (bad_alloc)
#endif
{
   void * ret = jalloc(s);
   if (ret == NULL) throw bad_alloc();
   return ret;
}

#ifdef WIN32
void * operator new[](size_t s)
#else
void * operator new[](size_t s) throw (bad_alloc)
#endif
{
   void * ret = jalloc(s);
   if (ret == NULL) throw bad_alloc();
   return ret;
}

#ifdef WIN32
void * operator new(size_t s, nothrow_t const &)
#else
void * operator new(size_t s, nothrow_t const &) throw ()
#endif
{
   return jalloc(s);
}

#ifdef WIN32
void * operator new[](size_t s, nothrow_t const &)
#else
void * operator new[](size_t s, nothrow_t const &) throw ()
#endif
{
   return jalloc(s);
}

#ifdef WIN32
void operator delete(void * p)
#else
void operator delete(void * p) throw()
#endif
{
   if (p) jfree(p);
}

#ifdef WIN32
void operator delete[](void * p)
#else
void operator delete[](void * p) throw()
#endif
{
   if (p) jfree(p);
}

#endif

status_t
ReflectServer ::
AddNewSession(AbstractReflectSessionRef ref, int socket)
{
   AbstractReflectSession * newSession = ref.GetItemPointer();
   if (newSession == NULL) return B_ERROR;

   newSession->_socket = socket;
   newSession->_owner = this;  // in case CreateGateway() needs to use the owner

   if (socket >= 0)
   {
      newSession->_gateway = newSession->CreateGateway(socket);
      if (newSession->_gateway == NULL) 
      {
         newSession->_socket = -1;
         return B_ERROR;
      }

      uint32 remoteIP;
      uint16 remotePort;
      if (GetPeerInfo(socket, remoteIP, remotePort) == B_NO_ERROR)
      {
         char ipBuf[32]; Inet_NtoA(remoteIP, ipBuf);
         newSession->_hostName = ipBuf;
         newSession->_port = remotePort;
      }
      else 
      {
         newSession->_hostName = "<unknown>";
         newSession->_port     = 0;
      }
   }
   else
   {
      newSession->_hostName = "<local>";
      newSession->_port     = 0;
   }

   return AddNewSession(ref);
}


status_t
ReflectServer ::
AddNewSession(AbstractReflectSessionRef ref)
{
   AbstractReflectSession * newSession = ref.GetItemPointer();
   if (newSession == NULL) return B_ERROR;
   newSession->_owner = this;

   if (_sessions.Put(newSession->GetSessionIDString(), ref))
   {
      if (newSession->AttachedToServer() == B_NO_ERROR)
      {
         LogTime(MUSCLE_LOG_INFO, "New %s [%s] from [%s:%i] (%lu total)\n", newSession->GetTypeName(), newSession->GetSessionIDString(), newSession->GetHostName(), newSession->GetClientPort(), _sessions.Size());
         return B_NO_ERROR;
      }
      else 
      {
         LogTime(MUSCLE_LOG_INFO, "%s [%s] from [%s:%i] aborted startup (%lu total)\n", newSession->GetTypeName(), newSession->GetSessionIDString(), newSession->GetHostName(), newSession->GetClientPort(), _sessions.Size());
         _sessions.Remove(newSession->GetSessionIDString());
      }
   }

   return B_ERROR;
}


static uint32 NullHashFunc(const short & key);
static uint32 NullHashFunc(const short & key)
{
   return (int32) key;
}

ReflectServer :: ReflectServer(uint32 maxBytes) : _portToAcceptor(NullHashFunc), _sessions(CStringHashFunc), _keepServerGoing(true), _startedAt(0)
{
   // make sure _lameDuckSessions has plenty of memory available in advance 
   // (as we need might need it in a tight spot later!)
   _lameDuckSessions.EnsureSize(256);

   _maxAllocableBytes = maxBytes;
   _nextPulseAt = MUSCLE_NEVER_PULSE;

   // This call lets the PortableMessage class use its default message pool for its
   // internal purposes.  It's safe to allow this here, since all PortableMessage calls
   // will be made from the same thread.  For a multi-threaded system, don't do this...
   PortableMessage::SetInternalDefaultPool(GetMessagePool());

   _sessions.SetCompareFunction(CStringCompareFunc, NULL, false);
}

ReflectServer :: ~ReflectServer()
{
   RemoveAcceptPort(-1);
}

void
ReflectServer :: Cleanup()
{
   // This silly little contortion is necessary to keep the
   // StorageReflectSession destructors from accessing themselves 
   // in the _sessions table while the _sessions table is being cleared!
   // (i.e. we keep an additional reference of each session around in
   // the (temp) array, so that we can clear _sessions without activating
   // all the destructors.  Then when we clear (temp), the destructors are
   // activated, but the _sessions table has already been safely cleared)
   PortableQueue<AbstractReflectSessionRef> temp(_sessions.Size());
   PortableHashtableIterator<const char *, AbstractReflectSessionRef> iter = GetSessions();
   AbstractReflectSessionRef next;
   while(iter.GetNextValue(next)) temp.AddTail(next);

   _sessions.Clear();
   _lameDuckSessions.MakeEmpty();
   temp.MakeEmpty();
}

status_t
ReflectServer ::
ReadyToRun()
{
   return B_NO_ERROR;
}

const char *
ReflectServer ::
GetServerName() const
{
   return "MUSCLE";
}

status_t 
ReflectServer ::
ServerProcessLoop()
{
   _startedAt = GetCurrentTime64();

   LogTime(MUSCLE_LOG_DEBUG, "This %s server was compiled on " __DATE__ " " __TIME__ "\n", GetServerName());
   LogTime(MUSCLE_LOG_DEBUG, "The server was compiled against MUSCLE version %s\n", MUSCLE_VERSION_STRING);
   if (ReadyToRun() != B_NO_ERROR) 
   {
      LogTime(MUSCLE_LOG_CRITICALERROR, "Server:  ReadyToRun() failed, aborting.\n");
      return B_ERROR;
   }

   // Print an informative startup message
   {
      int numFuncs = _portToAcceptor.Size();
      LogTime(MUSCLE_LOG_INFO, "%s Server is active", GetServerName()); 
      if (numFuncs > 0)
      {
         Log(MUSCLE_LOG_INFO, " and listening on port%s ", (numFuncs > 1) ? "s" : "");
         PortableHashtableIterator<short, ReflectAcceptor> iter = _portToAcceptor.GetIterator();
         short port;
         int which=0;
         while(iter.GetNextKey(port)) Log(MUSCLE_LOG_INFO, "%i%s", port, (++which<numFuncs)?",":"");
      }
      else Log(MUSCLE_LOG_INFO, " but not listening to any ports");
      Log(MUSCLE_LOG_INFO, ".\n");
   }

   fd_set readSet;
   fd_set writeSet;
   while(_keepServerGoing)
   {
      // Remove any sessions that were marked for removal
      // (Done first since this may have occurred even in ReadyToRun())
      {
         AbstractReflectSessionRef duckRef;
         while(_lameDuckSessions.RemoveHead(duckRef))
         {
            AbstractReflectSession * duck = duckRef.GetItemPointer();
            if (duck)
            {
               const char * id = duck->GetSessionIDString();
               if (_sessions.Remove(id)) LogTime(MUSCLE_LOG_INFO, "Closed %s [%s] to [%s:%i] (%lu total)\n", duck->GetTypeName(), id, duck->GetHostName(), duck->GetClientPort(), _sessions.Size());
            }
         }
      }

      FD_ZERO(&readSet);
      FD_ZERO(&writeSet);

      int maxSocket = -1;

      // Prepare the set with our acceptor sockets 
      {
         PortableHashtableIterator<short, ReflectAcceptor> iter = _portToAcceptor.GetIterator();
         ReflectAcceptor * next; 
         while((next = iter.GetNextValue()) != NULL)
         {
            int nextAcceptSocket = next->GetSocket();
            FD_SET(nextAcceptSocket, &readSet);
            if (nextAcceptSocket > maxSocket) maxSocket = nextAcceptSocket;
         }
      }

      // Find out which of our sessions wants to write data
      {
         const uint64 now64 = GetCurrentTime64();
         
         AbstractReflectSessionRef * nextRef;
         PortableHashtableIterator<const char *, AbstractReflectSessionRef> iter = GetSessions();
         while((nextRef = iter.GetNextValue()) != NULL)
         {
            AbstractReflectSession * session = nextRef->GetItemPointer();
            if ((session)&&(session->_gateway))
            {
               int sessionSocket = session->_socket;
               if (sessionSocket >= 0)
               {
                  bool in  = session->_gateway->GetInputEnabled(); 
                  bool out = session->_gateway->HasBytesToOutput();     
                  if (in)  FD_SET(sessionSocket, &readSet);
                  if (out) 
                  {
                     FD_SET(sessionSocket, &writeSet);
                     if (session->_lastByteOutputAt == 0) session->_lastByteOutputAt = now64;
                  }
                  else session->_lastByteOutputAt = 0;

                  if (((in)||(out))&&(sessionSocket > maxSocket)) maxSocket = sessionSocket;
               }
            }
         }
      }

      // Calculate timeout for Pulse(), if any
      struct timeval waitTime;
      {
         const uint64 now64 = GetCurrentTime64();
         _nextPulseAt = GetPulseTime(now64);

         // If we are now expecting a Pulse(), calculate how long to wait for it.
         if (_nextPulseAt != MUSCLE_NEVER_PULSE)
         {
            // Calculate how long to wait until the pulse time
            uint64 wt = (_nextPulseAt > now64) ? (_nextPulseAt - now64) : 0;
            Convert64ToTimeVal(wt, waitTime);
	    MASSERT((waitTime.tv_sec >= 0), "Negative select() timeout!");
         }
      }
 
      int numBitsSet = select(maxSocket+1, (maxSocket >= 0) ? &readSet : NULL, (maxSocket >= 0) ? &writeSet : NULL, NULL, (_nextPulseAt == MUSCLE_NEVER_PULSE) ? NULL : &waitTime);
      if (numBitsSet < 0) 
      {
         LogTime(MUSCLE_LOG_CRITICALERROR, "select() failed, aborting!\n");
         return B_ERROR;
      }

      if (_nextPulseAt != MUSCLE_NEVER_PULSE)
      {
         const uint64 now64 = GetCurrentTime64();
         if (now64 >= _nextPulseAt)
         {
            Pulse(now64);
            _nextPulseAt = MUSCLE_NEVER_PULSE;
         }
      }

      PortableHashtableIterator<short, ReflectAcceptor> iter = _portToAcceptor.GetIterator();
      ReflectAcceptor * acc;
      while((acc = iter.GetNextValue()) != NULL)
      {
         int acceptSocket = acc->GetSocket();
         if (FD_ISSET(acceptSocket, &readSet)) 
         {
            // Accept new connection and start up a session for it
            struct sockaddr_in saSocket;
            net_length_t nLen = sizeof(saSocket);
            int newSocket = accept(acceptSocket, (struct sockaddr *)&saSocket, &nLen);
            if (newSocket >= 0)
            {
               ReflectSessionFactory * factory = acc->GetFactory();
               AbstractReflectSession * newSession = factory ? factory->CreateSession() : NULL;
               if (newSession) 
               {
                  AbstractReflectSessionRef newSessionRef(newSession, NULL);
                  if (AddNewSession(newSessionRef, newSocket) != B_NO_ERROR) closesocket(newSocket);
               }
               else
               {
                  LogTime(MUSCLE_LOG_ERROR, "Error, couldn't create new reflector session!\n");
                  closesocket(newSocket);
               }
            }
            else LogTime(MUSCLE_LOG_WARNING, "Warning, accept() failed!\n");
         }
      }

      // Do the session I/Os...
      {
         const uint64 now = GetCurrentTime64();

         AbstractReflectSessionRef * nextRef;
         PortableHashtableIterator<const char *, AbstractReflectSessionRef> iter = GetSessions();
         while((nextRef = iter.GetNextValue()) != NULL)
         {
            AbstractReflectSession * session = nextRef->GetItemPointer();
            if ((session)&&(session->_gateway))
            {
               AbstractMessageIOGateway & gateway = *(session->_gateway);
               int socket = session->_socket;
               if (socket >= 0)
               {
                  uint32 readBytes = 0, wroteBytes = 0;
                  bool readError  = (FD_ISSET(socket, &readSet))&&(gateway.DoInput(readBytes)    != B_NO_ERROR);
                  bool writeError = (FD_ISSET(socket, &writeSet))&&(gateway.DoOutput(wroteBytes) != B_NO_ERROR); 
                  if ((readError)||(writeError)) 
                  {
                     if (session->ClientConnectionClosed()) AddLameDuckSession(*nextRef);
                     else
                     {
                        LogTime(MUSCLE_LOG_INFO, "TCP Connection to [%s:%i] was severed.\n", session->GetHostName(), session->GetClientPort());
                        session->_socket = -1;  // so we won't try to use it anymore
                     }
                  }
                  else if (session->_lastByteOutputAt > 0)
                  {
                     // Check for output stalls (new for v1.82)
                     if (wroteBytes > 0) session->_lastByteOutputAt = now;  // cool, he's streaming out
                     else if (now-session->_lastByteOutputAt > gateway.GetOutputStallLimit())
                     {
                        LogTime(MUSCLE_LOG_INFO, "TCP Connection to [%s:%i] timed out due to an output stall.\n", session->GetHostName(), session->GetClientPort());
                        if (session->ClientConnectionClosed()) AddLameDuckSession(*nextRef);
                                                          else session->_socket = -1;  // so we won't try to use it anymore
                     }
                  }
               }
            }
         }
      }

      // Let the various sessions handle their messages...
      while(true)
      {
         bool loopAgain = false;

         AbstractReflectSessionRef sRef;   // keep a copy here to prevent premature self-deletions
         PortableHashtableIterator<const char *, AbstractReflectSessionRef> iter = GetSessions();
         while((loopAgain == false)&&(iter.GetNextValue(sRef)))
         {
            AbstractReflectSession * s = sRef.GetItemPointer();

            if (s)
            {
               // Note:  It *is* necessary to check fromSession->_gateway every time through this
               //        loop, because a ReplaceSession() call might set it to NULL.
               PortableMessageRef nextMessageRef;
               _outOfMemoryWarningCount = 0;  // (s)'s responsibility for starts here!  If we run out of mem, he's history
               while((s->_gateway)&&(s->_gateway->GetNextIncomingMessage(nextMessageRef) == B_NO_ERROR)) 
               {
                  s->MessageReceivedFromGateway(nextMessageRef);
                  if (_outOfMemoryWarningCount > 0) break;  // big trouble!
                  if (s->_gateway == NULL)
                  {
                     // oops!  This session must have replaced itself.  Start through the loop again.
                     loopAgain = true;
                     break;
                  }
               }
               if (_outOfMemoryWarningCount > 0)
               {
                  _outOfMemoryWarningCount = 0;
                  LogTime(MUSCLE_LOG_CRITICALERROR, "Low Memory!  Aborting active session [%s:%i] to get some back!\n", s->GetHostName(), s->GetClientPort());
                  AddLameDuckSession(sRef);
                  DumpBoggedSessions();       // see what other cleanup we can do
                  break;
               }
            }
         }

         if (loopAgain == false) break;
      }
   }
   return B_NO_ERROR;
}


void ReflectServer :: DumpBoggedSessions()
{
   // New for v1.82:  also find anyone whose outgoing message queue is getting too large.
   //                 (where "too large", for now, is more than 5 megabytes)
   // This could happen if someone has a really slow Internet connection, or has decided to
   // stop reading from their end indefinitely for some reason.
   const int MAX_MEGABYTES = 5;
   AbstractReflectSessionRef * lRef;
   PortableHashtableIterator<const char *, AbstractReflectSessionRef> xiter = GetSessions();
   while((lRef = xiter.GetNextValue()) != NULL)
   {
      AbstractReflectSession * asr = lRef->GetItemPointer();
      if (asr)
      {
         AbstractMessageIOGateway * io = asr->GetGateway();
         if (io)
         {
            uint32 qSize = 0;
            const PortableQueue<PortableMessageRef> & q = io->GetOutgoingMessageQueue();
            for (int k=q.CountItems()-1; k>=0; k--)
            {
               const PortableMessage * qmsg = q[k].GetItemPointer();
               if (qmsg) qSize += qmsg->FlattenedSize();
            }
            if (qSize > MAX_MEGABYTES*1024*1024)
            {
               LogTime(MUSCLE_LOG_CRITICALERROR, "Low Memory!  Aborting bogged session [%s:%i] to get some back!\n", asr->GetHostName(), asr->GetClientPort());
               AddLameDuckSession(*lRef);
            }
         }
      }
   }
}

void 
ReflectServer ::
Pulse(uint64 now)
{
   AbstractReflectSessionRef * lRef; 
   PortableHashtableIterator<const char *, AbstractReflectSessionRef> xiter = GetSessions();
   while((lRef = xiter.GetNextValue()) != NULL)
   {
      AbstractReflectSession * session = lRef->GetItemPointer();
      if ((session)&&(session->_gateway))
      {
         AbstractMessageIOGateway * g = session->_gateway;
         if (now >= g->_nextPulseAt)
         {
            g->Pulse(now);
            g->_nextPulseAt = MUSCLE_NEVER_PULSE;
         }
      }
   }
}

uint64
ReflectServer ::
GetPulseTime(uint64 now)
{
   uint64 rt = MUSCLE_NEVER_PULSE;

   AbstractReflectSessionRef * lRef; 
   PortableHashtableIterator<const char *, AbstractReflectSessionRef> xiter = GetSessions();
   while((lRef = xiter.GetNextValue()) != NULL)
   {
      AbstractReflectSession * session = lRef->GetItemPointer();
      if ((session)&&(session->_gateway))
      {
         AbstractMessageIOGateway * g = session->_gateway;
         if (g->_nextPulseAt == MUSCLE_NEVER_PULSE) g->_nextPulseAt = g->GetPulseTime(now);
         if (g->_nextPulseAt < rt) rt = g->_nextPulseAt;
      }
   }
   return rt;
}


PortableHashtableIterator<const char *, AbstractReflectSessionRef>
ReflectServer ::
GetSessions() const
{
   return _sessions.GetIterator();
}

AbstractReflectSessionRef
ReflectServer ::
GetSession(const char * name) const
{
   AbstractReflectSessionRef ref;
   (void) _sessions.Get(name, ref); 
   return ref;
}

status_t
ReflectServer ::
ReplaceSession(AbstractReflectSessionRef newSessionRef, AbstractReflectSession * oldSession)
{
   // move the gateway from the old session to the new one...
   AbstractReflectSession * newSession = newSessionRef.GetItemPointer();
   if (newSession == NULL) return B_ERROR;

   newSession->_gateway  = oldSession->_gateway;
   newSession->_hostName = oldSession->_hostName;
   newSession->_port     = oldSession->_port;
   newSession->_socket   = oldSession->_socket;

   if (AddNewSession(newSessionRef) == B_NO_ERROR)
   {
      oldSession->_gateway = NULL;   /* gateway now belongs to newSession */
      EndSession(oldSession);
      return B_NO_ERROR;
   }
   else
   {
       // Oops, rollback changes and error out
       newSession->_gateway  = NULL;
       newSession->_hostName = "";
       newSession->_port     = 0; 
       newSession->_socket   = -1;
       return B_ERROR;
   }
}


void
ReflectServer ::
EndSession(AbstractReflectSession * who)
{
   AbstractReflectSessionRef * lRef; 
   PortableHashtableIterator<const char *, AbstractReflectSessionRef> xiter = GetSessions();
   while((lRef = xiter.GetNextValue()) != NULL)
   {
      AbstractReflectSession * session = lRef->GetItemPointer();
      if (session == who)
      {
         AddLameDuckSession(*lRef);
         return;
      }
   }
}

void
ReflectServer ::
EndServer()
{
   _keepServerGoing = false;
}

status_t
ReflectServer ::
PutAcceptPort(ReflectSessionFactory * factory, short port)
{
   if (port > 0)
   {
      int acceptSocket = CreateAcceptingSocket(port);
      if (acceptSocket >= 0)
      {
         // Get rid of any previous acceptor on this port...
         RemoveAcceptPort(port);

         // And add the new one
         if (_portToAcceptor.Put(port, ReflectAcceptor(factory, acceptSocket))) return B_NO_ERROR;

         closesocket(acceptSocket);
      }
   }
   return B_ERROR;
}

void
ReflectServer ::
RemoveAcceptPort(short port)
{
   if (port >= 0)
   {
      ReflectAcceptor acc;
      if (_portToAcceptor.Remove(port, acc)) 
      {
         closesocket(acc.GetSocket());
         delete acc.GetFactory();
      }
   }
   else
   {
      PortableHashtableIterator<short, ReflectAcceptor> iter = _portToAcceptor.GetIterator();
      ReflectAcceptor next;
      while(iter.GetNextValue(next)) 
      {
         closesocket(next.GetSocket());
         delete next.GetFactory();
      }
      _portToAcceptor.Clear();
   }
}


uint64
ReflectServer ::
GetServerUptime() const
{
   return (_startedAt > 0) ? (GetCurrentTime64() - _startedAt) : 0;
}
                        
void
ReflectServer :: AddLameDuckSession(AbstractReflectSessionRef ref)
{
   if ((_lameDuckSessions.IndexOf(ref) < 0)&&(_lameDuckSessions.AddTail(ref) == false)) LogTime(MUSCLE_LOG_CRITICALERROR, "Server:  AddLameDuckSession() failed, I'm REALLY in trouble!  Aggh!\n");
}
