/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef STORAGE_REFLECT_SESSION_H
#define STORAGE_REFLECT_SESSION_H

#include "reflector/DumbReflectSession.h"
#include "reflector/StorageReflectConstants.h"
#include "regex/PathMatcher.h"

/**
 *  This is a factory class that returns new StorageReflectSession objects.
 */
class MUSCLE_CLASS_DECLARATIONS StorageReflectSessionFactory : public ReflectSessionFactory
{
public:
   virtual AbstractReflectSession * CreateSession();
};       

/** An intelligent AbstractReflectSession that knows how to
 *  store data on the server, and filter using wildcards.  This class
 *  is used by the muscled server program to handle incoming connections.
 *  See StorageReflectConstants.h and/or "The Beginner's Guide.html"
 *  for details.
 */
class MUSCLE_CLASS_DECLARATIONS StorageReflectSession : public DumbReflectSession
{
public:
   /** Default constructor. */
   StorageReflectSession();

   /** Virtual Destructor. */
   virtual ~StorageReflectSession();

   /** Called after the constructor, when the session is ready to interact with the server.
    *  @return B_NO_ERROR if everything is okay to go, B_ERROR if there was a problem
    *          setting up, or if the IP address of our client has been banned. 
    */
   virtual status_t AttachedToServer();

   /** Called when a new message is received from our IO gateway. */
   virtual void MessageReceivedFromGateway(PortableMessageRef msg);

   /** Drains the storage pools used by StorageReflectSessions, to recover memory */
   static void DrainPools();

protected:
   /**
    * Create or Set the value of a data node.
    * @param nodePath Should be the path relative to the home dir (e.g. "MyNode/Child1/Grandchild2")
    * @param dataMsgRef The value to set the node to
    * @param allowOverwriteData Indicates whether existing node-data may be overwritten.  If false, the method will fail if the specified node already exists.
    * @param allowCreateNode indicates whether new nodes may be created.  (If false,
    * the method will fail if any node in the specified node path 
    * doesn't already exist)
    * @param quiet If set to true, subscribers won't be updated regarding this change to the database.
    * @return B_NO_ERROR on success, or B_ERROR on failure.
    */
   status_t SetDataNode(const PortableString & nodePath, PortableMessageRef dataMsgRef, bool allowOverwriteData=true, bool allowCreateNode=true, bool quiet=false);

   /** Returns a human-readable label for this session type:  "Storage Reflect Session" */
   virtual const char * GetTypeName() const {return "Storage Reflect Session";}

   class DataNode;

   /** Type for a Reference to a DataNode object */
   typedef PortableRef<DataNode> DataNodeRef;

   /** Iterator type for our child objects */
   typedef PortableHashtableIterator<const char *, DataNodeRef> DataNodeRefIterator;

   /** Each object of this class represents one node in the server-side data-storage tree.  */
   class DataNode : public PortableRefCountable
   {
   public:
      /**
       * Put a child without changing the ordering index
       * @param child Reference to the child to accept into our list of children
       * @param optNotifyWithOnSetParent If non-NULL, a StorageReflectSession to use to notify subscribers that the new node has been added
       * @param optNotifyWithOnChangedData If non-NULL, a StorageReflectSession to use to notify subscribers when the node's data has been alterered
       * @return true on success, false if out of memory
       */
      bool PutChild(DataNodeRef & child, StorageReflectSession * optNotifyWithOnSetParent, StorageReflectSession * optNotifyWithOnChangedData);
 
      /**
       * Create and add a new child node for (data), and put it into the ordering index
       * @param data Reference to a message to create a new child node for.
       * @param optInsertBefore if non-NULL, the name of the child to put the new child before in our index.  If NULL, (or the specified child cannot be found) the new child will be appended to the end of the index.
       * @param optNotifyWithOnSetParent If non-NULL, a StorageReflectSession to use to notify subscribers that the new node has been added
       * @param optNotifyWithOnChangedData If non-NULL, a StorageReflectSession to use to notify subscribers when the node's data has been alterered
       * @return true on success, false if out of memory
       */
      bool InsertOrderedChild(PortableMessageRef data, const char * optInsertBefore, StorageReflectSession * optNotifyWithOnSetParent, StorageReflectSession * optNotifyWithOnChangedData);
 
      /** 
       * Moves the given node (which must be a child of ours) to be just before the node named
       * (moveToBeforeThis) in our index.  If (moveToBeforeThis) is not a node in our index,
       * then (child) will be moved back to the end of the index. 
       * @param child A child node of ours, to be moved in the node ordering index.
       * @param moveToBeforeThis name of another child node of ours.  If this name is NULL or
       *                         not found in our index, we'll move (child) to the end of the index.
       * @param optNotifyWith If non-NULL, this will be used to sent INDEXUPDATE message to the
       *                      interested clients, notifying them of the change.
       */
      bool ReorderChild(const DataNode & child, const char * moveToBeforeThis, StorageReflectSession * optNotifyWith);

      /** Returns true iff we have a child with the given name */
      bool HasChild(const char * key) const {return ((_children)&&(_children->ContainsKey(key)));}

      /** Retrieves the child with the given name.
       *  @param key The name of the child we wish to retrieve
       *  @param returnChild On success, a reference to the retrieved child is written into this object.
       *  @return B_NO_ERROR if a child node was successfully retrieved, or B_ERROR if it was not found.
       */
      status_t GetChild(const char * key, DataNodeRef & returnChild) const {return ((_children)&&(_children->Get(key, returnChild))) ? B_NO_ERROR : B_ERROR;}

      /** Removes the child with the given name.
       *  @param key The name of the child we wish to remove.
       *  @param optNotifyWith If non-NULL, the StorageReflectSession that should be used to notify subscribers that the given node has been removed
       *  @param recurse if true, the removed child's children will be removed from it, and so on, and so on...
       *  @param optCounter if non-NULL, this value will be incremented whenever a child is removed (recursively)
       *  @return B_NO_ERROR if the given child is found and removed, or B_ERROR if it could not be found.
       */
      status_t RemoveChild(const char * key, StorageReflectSession * optNotifyWith, bool recurse, uint32 * optCounter);

      /** Returns an iterator that can be used for iterating over our set of children. */
      DataNodeRefIterator GetChildIterator() const {return _children ? _children->GetIterator() : DataNodeRefIterator();}

      /** Returns the number of child nodes this node contains. */
      uint32 CountChildren() const {return _children ? _children->Size() : 0;}

      /** Returns the ASCII name of this node (e.g. "joe") */
      const char * GetNodeName() const {return _nodeName.Cstr();}

      /** Returns the full node path of this node (e.g. "/12.18.240.15/1234/beshare/files/joe") */
      const char * GetNodePath() const {return _nodePath.Cstr();}

      /** Replaces this node's payload message with that of (data).
       *  @data the new PortableMessage to associate with this node.
       *  @param optNotifyWith if non-NULL, this StorageReflectSession will be used to notify subscribers that this node's data has changed.
       */
      void SetData(PortableMessageRef data, StorageReflectSession * optNotifyWith);

      /** Returns a reference to this node's PortableMessage payload. */
      PortableMessageRef GetData() const {return _data;}

      /** Returns our node's parent, or NULL if this node doesn't have a parent node. */
      DataNode * GetParent() const {return _parent;}

      /** Returns this node's depth in the tree (e.g. zero if we are the root node, 1 if we are its child, etc) */
      int GetDepth() const {return _depth;}

      /** Returns us to our virgin, pre-Init() state, by clearing all our children, subscribers, parent, etc.  */
      void Reset();  

      /**
       * Modifies the refcount for the given sessionID.
       * Any sessionID's with (refCount > 0) will be in the GetSubscribers() list.
       * @param sessionID the sessionID whose reference count is to be modified
       * @param delta the amount to add to the reference count.
       */
      void IncrementSubscriptionRefCount(const char * sessionID, long delta);

      /** Returns an iterator that can be used to iterate over our list of active subscribers */
      PortableHashtableIterator<const char *, uint32> GetSubscribers() const;

      /** Returns a reference to our ordered-child index */
      const PortableQueue<const char *> * GetIndex() const {return _orderedIndex;}

      /** Default Constructor.  Don't use this, use StorageReflectSession::GetNewDataNode() instead!  */
      DataNode();

      /** Default Destructor.   Don't use this, use StorageReflectSession::ReleaseDataNode() instead!  */
      ~DataNode();

   private:
      friend class StorageReflectSession;

      void Init(const char * nodeName, PortableMessageRef initialValue);
      void SetParent(DataNode * _parent, StorageReflectSession * optNotifyWith);
      bool RemoveIndexEntry(const char * key, StorageReflectSession * optNotifyWith);

      DataNode * _parent;
      PortableMessageRef _data;
      PortableHashtable<const char *, DataNodeRef> * _children;  // lazy-allocated
      PortableQueue<const char *> * _orderedIndex;  // only used when tracking the ordering of our children (lazy-allocated)
      uint32 _orderedCounter;
      PortableString _nodeName;
      PortableString _nodePath;
      int _depth;  // number of ancestors our node has (e.g. root's _depth is zero)

      PortableHashtable<const char *, uint32> _subscribers; 
   };

   /**
    * This typedef represents the proper signature of a node-tree traversal callback function.
    * Functions with this signature may be used with the NodePathMatcher::DoTraversal() method.
    * @param This The StorageReflectSession doing the traveral, as specified in the DoTraversal() call
    * @param node The DataNode that was matched the criteria of the traversal
    * @param userData the same value that was passed in to the DoTraversal() method.
    * @return The callback function should return the depth at which the traversal should continue after
    *         the callback is done.  So to allow the traversal to continue normally, return (node.GetDepth()),
    *         or to terminate the traversal immediately, return 0, or to resume the search at the next
    *         session, return 2, or etc.
    */
   typedef int (*PathMatchCallback)(StorageReflectSession * This, DataNode & node, void * userData);

   /** A slightly extended version of PathMatcher that knows how to handle DataNodes directly. */
   class NodePathMatcher : public PathMatcher
   {
   public:
      /** Default constructor. */
      NodePathMatcher()  {/* empty */}

      /** Destructor. */
      ~NodePathMatcher() {/* empty */}

      /**
       * Returns true iff the given node matches our query.
       * @param node the node to check to see if it matches
       * @param rootDepth the depth at which the traversal started (i.e. 0 if started at root)
       */
      bool MatchesNode(DataNode & node, int rootDepth) const;

      /**
       * Does a depth-first traversal of the node tree, starting with (node) as the root.
       * @param cb The callback function to call whenever a node is encountered in the traversal
       *           that matches at least one of our path strings.
       * @param node The node to begin the traversal at.
       * @param userData Any value you wish; it will be passed along to the callback method.
       */
      void DoTraversal(PathMatchCallback cb, StorageReflectSession * This, DataNode & node, void * userData);
 
      /**
       * Returns the number of path-strings that we contain that match (node).
       * Note this is a bit more expensive than MatchesNode(), as we can't use short-circuit boolean logic here!
       * @param node A node to check against our set of path-matcher strings.
       * @param rootDepth the depth at which the traversal started (i.e. 0 if started at root)
       */
      int GetMatchCount(DataNode & node, int nodeDepth) const;

   private:
      int DoTraversalAux(PathMatchCallback cb, StorageReflectSession * This, DataNode & node, void * userData, int rootDepth);
      bool PathMatches(DataNode & node, int whichPath, int rootDepth) const;
   };

   friend class StorageReflectSession::DataNode;

   /**
    * Returns true iff we have the given PR_PRIVILEGE_* privilege.
    * @return Default implementation looks at the PR_NAME_PRIVILEGE_BITS parameter.
    */
   virtual bool HasPrivilege(int whichPriv) const;

   /**
    * Returns the given PortableMessage to our client, inside an error message
    * with the given error code.
    */
   void BounceMessage(uint32 errorCode, PortableMessageRef msgRef);

   /** Removes our nodes from the tree;  called by the destructor or on setup errors.  */
   void Cleanup();

   /** Convenience method (used by some customized daemons) -- Given a source node and a destination path,
     * Make (path) a deep, recursive clone of (node).  (note that node data messages are not cloned, however)
     * @param sourceNode Reference to a DataNode to clone.
     * @param destPath Path of where the newly created node subtree will appear.  Should be relative to our home node.
     * @return B_NO_ERROR on success, or B_ERROR on failure (may leave a partially cloned subtree on failure)
     */
   status_t CloneDataNodeSubtree(const StorageReflectSession::DataNode & sourceNode, const PortableString & destPath);

   /** Tells other sessions that we have modified (node).  */
   virtual void NotifySubscribersThatNodeChanged(DataNode & node, bool isBeingRemoved);

   /** Tells other sessions that we have a new (node) available.  */
   virtual void NotifySubscribersOfNewNode(DataNode & newNode);

   /** Tells other sessions that we have changed the index of (node).  */
   virtual void NotifySubscribersThatNodeIndexChanged(DataNode & node, char op, uint32 index, const char * key);

   /** Called by NotifySubscribersOfNewNode to tell us that (node) has been modified.  */
   void NodeChanged(DataNode & node, bool isBeingRemoved);

   /** Called by NotifySubscribersThatIndexChanged to tell us how (node)'s index has been modified.  */
   void NodeIndexChanged(DataNode & node, char op, uint32 index, const char * key);

   /**
    * Called by SetParent() to tell us that (node) has been created at a given location.
    * We then respond by letting any matching subscriptions add their mark to the node.
    * @param node the new Node that has been added to the database.
    */
   void NodeCreated(DataNode & node);

   /**
    * Takes any messages that were created in the NodeChanged() callbacks and 
    * sends them to their owner's MessageReceivedFromNeighbor() method for
    * processing and eventual forwarding to the client.
    */
   void PushSubscriptionMessages();

   /** Auxilliary helper method for PushSubscriptionMessages() */
   void PushSubscriptionMessage(PortableMessage ** msgField); 

   /**
    * Executes a data-gathering tree traversal based on the PR_NAME_KEYS specified in the given message.
    * @param getMsg a PortableMessage whose PR_NAME_KEYS field specifies which nodes we are interested in.
    */
   void DoGetData(const PortableMessage & getMsg);

   /** Auxilliary helper function. */
   void SendGetDataResults(PortableMessage ** msg);

   /**
    * If set false, we won't receive subscription updates.
    * @param e Whether or not we wish to get update messages from our subscriptions.
    */
   void SetSubscriptionsEnabled(bool e) {_subscriptionsEnabled = e;}

   /** Returns true iff our "subscriptions enabled" flag is set.  Default state is of this flag is true.  */
   bool GetSubscriptionsEnabled() const {return _subscriptionsEnabled;}

   /**
    * Convenience method:  Uses the given path to lookup a single node in the node tree
    * and return it.  (Note that wildcards are not supported by this method!)
    * If (path) begins with a '/', the search will begin with the root
    * node of the tree; if not, it will begin with this session's node.  Returns NULL on failure.
    * @param path The fully specified path to a single node in the database.
    * @return A pointer to the specified DataNode, or NULL if the node wasn't found.
    */
   DataNode * GetDataNode(const PortableString & path) const;

   /** Traversal callback:  matching nodes are ref'd/unref'd with subscribed session IDs */
   int DoSubscribeRefCallback(DataNode & node, void * userData);

   /** object-pool/recycling callback */
   static void ResetMatcherQueueFunc(StringMatcherQueue * q, void * /*userData*/) {q->MakeEmpty();}

   /** object-pool/recycling callback */
   static void ResetNodeFunc(StorageReflectSession::DataNode * node, void * /*userData*/) {node->Reset();}

   /** callback stub */
   static int DoSubscribeRefCallbackFunc(StorageReflectSession * This, DataNode & node, void * userData) {return This->DoSubscribeRefCallback(node, userData);}

   /** Traversal callback:  matching nodes are sent the given message (userData).  */
   int PassMessageCallback(DataNode & node, void * userData);

   /** callback stub */
   static int PassMessageCallbackFunc(StorageReflectSession * This, DataNode & node, void * userData) {return This->PassMessageCallback(node, userData);}

   /** Traversal callback:  owners of matching nodes are EndSession()'d (i.e. kicked off the server) */
   int KickClientCallback(DataNode & node, void * userData);

   /** callback stub */
   static int KickClientCallbackFunc(StorageReflectSession * This, DataNode & node, void * userData) {return This->KickClientCallback(node, userData);}

   /** Traversal callback:  matching nodes are placed into the given message (userData).  */
   int GetDataCallback(DataNode & node, void * userData);

   /** callback stub */
   static int GetDataCallbackFunc(StorageReflectSession * This, DataNode & node, void * userData) {return This->GetDataCallback(node, userData);}

   /** Traversal callback:  matching nodes are placed in a list (userData) to be removed from the node tree. */
   int RemoveDataCallback(DataNode & node, void * userData);

   /** callback stub */
   static int RemoveDataCallbackFunc(StorageReflectSession * This, DataNode & node, void * userData) {return This->RemoveDataCallback(node, userData);}

   /** Traversal callback:  matching nodes have ordered data inserted into them as children.  */
   int InsertOrderedDataCallback(DataNode & node, void * userData);

   /** callback stub */
   static int InsertOrderedDataCallbackFunc(StorageReflectSession * This, DataNode & node, void * userData) {return This->InsertOrderedDataCallback(node, userData);}

   /** Traversal callback:  matching nodes are reordered in their parent's index */
   int ReorderDataCallback(DataNode & node, void * userData);

   /** callback stub */
   static int ReorderDataCallbackFunc(StorageReflectSession * This, DataNode & node, void * userData) {return This->ReorderDataCallback(node, userData);}

   /**
    * Call this to get a new DataNode, instead of using the DataNode ctor directly.
    * @param nodeName The name to be given to the new DataNode that will be created.
    * @param initialValue The PortableMessage payload to be given to the new DataNode that will be created.
    * @return A pointer to the new DataNode (caller assumes responsibility for it), or NULL if out of memory.
    */
   DataNode * GetNewDataNode(const char * nodeName, PortableMessageRef initialValue);

   /**
    * Call this when you are done with a DataNode, instead of the DataNode destructor.
    * @param node The DataNode to delete or recycle.
    */
   void ReleaseDataNode(DataNode * node);

   /**
    * This method goes through the outgoing-messages list looking for PR_RESULT_DATAITEMS
    * messages.  For each such message that it finds, it will remove all result items
    * (both update items and items in PR_NAME_REMOVED_DATAITEMS) that match the expressions
    * in the given NodePathMatcher.  If the NodePathMatcher argument is NULL, all results
    * will be removed.  Any PR_RESULT_DATAITEMS messages that become empty will be removed
    * completely from the queue.
    * @param matcher Specifies which data items to delete from messages in the outgoing message queue.
    */
   void JettisonOutgoingResults(const NodePathMatcher * matcher);

   /** our current parameter set */
   PortableMessage _parameters;  

   /** cached to be sent when a subscription triggers */
   PortableMessage * _nextSubscriptionMessage;  

   /** cached to be sent when an index subscription triggers */
   PortableMessage * _nextIndexSubscriptionMessage;  

   /** Points to shared global root node; same for all sessions */
   DataNode * _globalRoot;       

   /** this session's subdir (grandchild of _globalRoot) */
   DataNodeRef _sessionDir;      

   /** Our session's set of active subscriptions */
   NodePathMatcher _subscriptions;  

   /** Where user messages get sent if no PR_NAME_KEYS field is present */
   NodePathMatcher _defaultMessageRoute;  

   /** Whether or not we set to report subscription updates or not */
   bool _subscriptionsEnabled;            

   /** Maximum number of subscription update fields per PR_RESULT message */
   int32 _maxSubscriptionMessageItems;    

   /** Optimization flag:  set true the first time we index a node */
   bool _indexingPresent;                 

   /** The number of database nodes we currently have created */
   uint32 _currentNodeCount;              

   /** The maximum number of database nodes we are allowed to create */
   uint32 _maxNodeCount;                  

   /** Our pool of saved database nodes */
   static DataNodeRef::ItemPool _nodePool;   

   /** Our pool of saved StringMatchers */
   static StringMatcherRef::ItemPool _stringMatcherPool;   

   /** Our pool of saved StringMatcherQueues */
   static StringMatcherQueueRef::ItemPool _stringMatcherQueuePool;    

   friend class StorageReflectSession :: NodePathMatcher;
};

#endif
