/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <time.h>

#include "dataio/TCPSocketDataIO.h"
#include "iogateway/PortableMessageIOGateway.h"
#include "reflector/StorageReflectConstants.h"
#include "util/NetworkUtilityFunctions.h"

static void Kick(PortableMessageIOGateway & gw, const char * arg);
void Kick(PortableMessageIOGateway & gw, const char * arg)
{
   PortableMessage * msg = GetMessageFromPool(PR_COMMAND_KICK);
   PortableString str("/");
   str += arg;
   str += "/*";
   msg->AddString(PR_NAME_KEYS, str);
   gw.AddOutgoingMessage(PortableMessageRef(msg, GetMessagePool()));
   LogTime(MUSCLE_LOG_INFO, "Kicking users matching pattern [%s]\n", str.Cstr());
}

static void Ban(PortableMessageIOGateway & gw, const char * arg, bool unBan);
void Ban(PortableMessageIOGateway & gw, const char * arg, bool unBan)
{
   PortableMessage * msg = GetMessageFromPool(unBan ? PR_COMMAND_REMOVEBANS : PR_COMMAND_ADDBANS);
   msg->AddString(PR_NAME_KEYS, arg);
   gw.AddOutgoingMessage(PortableMessageRef(msg, GetMessagePool()));
   if (unBan) LogTime(MUSCLE_LOG_INFO, "Removing ban patterns that match pattern [%s]\n", arg);
         else LogTime(MUSCLE_LOG_INFO, "Adding ban pattern [%s]\n", arg);
}


// This is a little admin program, useful for kicking, banning, or unbanning users
// without having to restart the MUSCLE server.  Example command line:
//   admin server=muscleserver.mycompany.com kick=192.168.0.23 ban=16.25.29.2 kickban=1.2.3.4 unban=1.2.3.4 ban=2.3.4.5 ban=3.4.5.*
// Note that you can only do this if your IP address has the requisite privileges on
// the MUSCLE server!  (i.e. to ban, the server must have been run with an argument like privban=your.ip.address)
int main(int argc, char ** argv)
{
   if (InitializeTCPStack() != B_NO_ERROR) {LogTime(MUSCLE_LOG_CRITICALERROR, "%s couldn't initialize TCP stack!\n", argv[0]); return 10;}

   char * hostName = "localhost";

   // First, find out if there is a server specified.
   for (int i=1; i<argc; i++)
   {
      char * next = argv[i];
      if (strncmp(next, "server=", 7) == 0) hostName = &next[7];
      else if (strcmp(next, "help") == 0)
      {
         LogTime(MUSCLE_LOG_INFO, "This program lets you send admin commands to a running MUSCLE server.\n");
         LogTime(MUSCLE_LOG_INFO, "Note that the MUSCLE server won't listen to your commands unless your ip address was\n");
         LogTime(MUSCLE_LOG_INFO, "specified as a privileged IP address in its command line arguments [e.g. ./muscled privall=your.IP.address]\n");
         LogTime(MUSCLE_LOG_INFO, "Usage:  admin [server=localhost] [ban=pattern] [unban=pattern] [kick=pattern] [kickban=pattern]\n");
         exit(0);
      }
   }
   
   char * colon = strchr(hostName, ':');
   int port = (colon) ? atoi(colon+1) : 2960;
   if (colon) *colon = '\0';

   int s = Connect(hostName, port, "admin", false);
   if (s < 0) 
   {
      LogTime(MUSCLE_LOG_INFO, "(run 'admin help' for arguments)\n");
      exit(10);
   }

   PortableMessageIOGateway gw(new TCPSocketDataIO(s, false));

   // Now generate all commands...
   for (int j=1; j<argc; j++)
   {
      char * line = argv[j];
      char * colon = strchr(line, '=');
      if (colon)
      {
         char * arg = colon+1;
         *colon = '\0';
              if (strcmp(line, "kick")    == 0) Kick(gw, arg);
         else if (strcmp(line, "ban")     == 0) Ban(gw, arg, false);
         else if (strcmp(line, "unban")   == 0) Ban(gw, arg, true);
         else if (strcmp(line, "kickban") == 0)
         {
            Kick(gw, arg);
            Ban(gw, arg, false);
         }
      }
   }

   // Lastly, request a PONG so we know when all has been done
   gw.AddOutgoingMessage(PortableMessageRef(GetMessageFromPool(PR_COMMAND_PING), GetMessagePool()));

   // send them, and then wait for the pong back
   uint32 errorCount = 0;
   bool keepGoing = true;
   while(keepGoing)
   {
      fd_set readSet, writeSet;
      FD_ZERO(&readSet);
      FD_ZERO(&writeSet);

      FD_SET(s, &readSet);
      if (gw.HasBytesToOutput()) FD_SET(s, &writeSet);

      if (select(s+1, &readSet, &writeSet, NULL, NULL) < 0)
      {
         LogTime(MUSCLE_LOG_CRITICALERROR, "select() failed, exiting!\n");
         errorCount++;
         break;
      }

      uint32 junk = 0;
      bool readError  = (FD_ISSET(s, &readSet))&&(gw.DoInput(junk) != B_NO_ERROR);
      bool writeError = (FD_ISSET(s, &writeSet))&&(gw.DoOutput(junk) != B_NO_ERROR);
      if ((readError)||(writeError))
      {
         LogTime(MUSCLE_LOG_ERROR, "TCP connection was cut prematurely!\n");
         errorCount++;
         break;
      }

      PortableMessageRef incoming;
      while(gw.GetNextIncomingMessage(incoming) == B_NO_ERROR)
      {
         const PortableMessage * msg = incoming();
         if (msg)
         {
            switch(msg->what)
            {
               case PR_RESULT_PONG:
                  keepGoing = false;
               break;

               case PR_RESULT_ERRORACCESSDENIED:
               {
                  errorCount++;
                  PortableMessageRef subMsg;
                  const char * who;
                  LogTime(MUSCLE_LOG_ERROR, "Access denied!  ");
                  if ((msg->FindMessage(PR_NAME_REJECTED_MESSAGE, subMsg) == B_NO_ERROR)&&(subMsg())&&(subMsg()->FindString(PR_NAME_KEYS, &who) == B_NO_ERROR)) Log(MUSCLE_LOG_ERROR, "You aren't allowed to %s [%s]!", (subMsg()->what == PR_COMMAND_KICK) ? "kick" : ((subMsg()->what == PR_COMMAND_ADDBANS) ? "ban" : "unban"), who); 
                  Log(MUSCLE_LOG_ERROR, "\n");
               }
               break;
            }
         }
      }     
   }
   LogTime(MUSCLE_LOG_INFO, "Exiting. (%lu errors)\n", errorCount);
   return 0;
}
