/* This file is Copyright 2001 Level Control Systems.  See the included LICENSE.txt file for details. */

#ifndef OBJECT_POOL_H
#define OBJECT_POOL_H

#include "support/PortableTypeDefs.h"
#include "util/PortableQueue.h"

// Uncomment this #define to disable object pools (i.e. turn them back
// into fancy new/delete operations).  This is helpful if you are trying
// to track down memory leaks.
//#define DISABLE_OBJECT_POOLING 1

/** A templated class that helps reduce the number of dynamic allocations
 *  a deallocations in your app.  Instead of calling new Object, you call
 *  myObjectPool.GetObject(), and instead of calling delete Object, you call
 *  myObjectPool.ReleaseObject().  The advantage is that the ObjectPool will
 *  keep (up to) a certain number of "spare" Objects around, and recycle them back
 *  to you as needed. 
 */
template <class Object> class MUSCLE_CLASS_DECLARATIONS ObjectPool
{
public:
   /** Callback function type for Init and Recycle callbacks */
   typedef void (*ObjectCallback)(Object * obj, void * userData);

   /**
    *  Constructor.
    *  @param maxPoolSize the maximum number of recycled objects that may be kept around for future reuse at any one time.
    *  @param recycleCallback optional callback function to be called whenever an object is returned to the pool.
    *  @param recycleData user value to be passed in to the recycleCallback function.
    *  @param initCallback optional callback function to be called whenever an object is taken out of the pool.
    *  @param initCallbackData user value to be passed in to the initCallback function.
    */      
   ObjectPool(uint32 maxPoolSize=100, 
              ObjectCallback recycleCallback = NULL, void * recycleData = NULL,
              ObjectCallback initCallback = NULL, void * initData = NULL) : 
      _initObjectFunc(initCallback), _initObjectUserData(initData),
      _recycleObjectFunc(recycleCallback), _recycleObjectUserData(recycleData),
      _maxPoolSize(maxPoolSize)
   {
      // empty
   }

   /** 
    *  Destructor.
    *  Deletes all held "spare" objects in the pool.
    */
   virtual ~ObjectPool()
   {
      Drain();
   }

   /** Returns a new Object for use (or NULL if no memory available).
    *  You are responsible for calling ReleaseObject() on this object
    *  when you are done with it.  This method can be used interchangably
    *  with 'new (nothrow) Object()'.
    *  @return a new Object, or NULL if out of memory.
    */
   Object * GetObject()
   {
#ifdef DISABLE_OBJECT_POOLING
      Object * ret = newnothrow Object;
      if (ret == NULL) WARN_OUT_OF_MEMORY;
      return ret;
#else
      Object * ret = NULL;

      Lock();
      (void) _pool.RemoveTail(ret);
      Unlock();

      if (ret == NULL) ret = newnothrow Object;
      if (ret)
      {
         if (_initObjectFunc) _initObjectFunc(ret, _initObjectUserData);
      }
      else WARN_OUT_OF_MEMORY;
      return ret;
#endif
   }

   /** Adds the given object to our "standby" object list to be recycled, or
    *  deletes it if the "standby" list is already at its maximum size.
    *  @param obj An Object to recycle or delete.  May be NULL.  
    *             This method can be used interchangable with 'delete Object'.
    *  @param numItems Number of Objects pointed to by (obj).  In most cases,
    *                  this will be 1, but if (obj) is pointing to an array,
    *                  this argument should be set to the length of the array.
    *                  (or at least to some number other than one)
    */
   void ReleaseObject(Object * obj, uint32 numItems = 1)
   {
#ifdef DISABLE_OBJECT_POOLING
      if (numItems == 1) delete obj; else delete [] obj;  // Under mwcc, this delete doesn't call the dtor!?  WTF??  --jaf
#else
      if (obj)
      {
         if (_recycleObjectFunc) _recycleObjectFunc(obj, _recycleObjectUserData);
         Lock();
         if (_pool.CountItems() < _maxPoolSize)
         {
            if (_pool.AddTail(obj) == false) delete obj;
         }
         else 
         {
            if (numItems == 1) delete obj;
                          else delete [] obj;
         }
         Unlock();
      }
#endif
   }

  /**
   * With this call you can specify a callback function that
   * will be called whenever a new Object is created by this pool. 
   * If you need to do any initialization on the new Object, you can
   * do it here.
   * @param cb The callback function that GetObject() will call.
   * @param userData User value that will be passed through to the callback function.
   */
   void SetInitObjectCallback(ObjectCallback cb, void * userData)
   {
      _initObjectFunc = cb;
      _initObjectUserData = userData;
   }

  /**
   * With this call you can specify a callback function that
   * will be called by ReleaseObject() whenever it is putting an
   * Object into its standby list.
   * If you wish to do any cleanup on the object, you can do it here.
   * @param cb The callback function that ReleaseObject() will call.
   * @param userData User value that will be passed through to the callback function.
   */
   void SetRecycleObjectCallback(ObjectCallback cb, void * userData)
   {
      _recycleObjectFunc = cb;
      _recycleObjectUserData = userData;
   }

  /** Removes all "spare" objects from the pool and deletes them. */
   void Drain()
   {
      for (int i=_pool.CountItems()-1; i>=0; i--) delete _pool[i];
      _pool.MakeEmpty();
   }

protected:
   /** This method is a no-op, but may be overridden in subclasses to lock
    *  this ObjectPool in order to make it thread-safe.
    */
   virtual void Lock() {/* empty:  subclasses may override this to provide thread safety */}

   /** This method is a no-op, but may be overridden in subclasses to unlock
    *  this ObjectPool as part of making it thread-safe.
    */
   virtual void Unlock() {/* empty:  subclasses may override this to provide thread safety */}

private:
   ObjectCallback _initObjectFunc;
   void * _initObjectUserData;
   
   ObjectCallback _recycleObjectFunc;
   void * _recycleObjectUserData;
   
   PortableQueue<Object *> _pool;
   uint32 _maxPoolSize;
};

#endif
